# -*- coding: utf-8 -*-
"""
NCBImeta Error Classes

@author: Katherine Eaton
"""

class ErrorOutputDirNotExists(Exception):
    '''Error class for when an output directory does not exist'''
    def __init__(self, value):
        '''
        The constructor for ErrorOutputDirNotExists class.

        Parameters:
        value(str): Path to the output directory.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the output directory path.'''
        return ("\n\nOutput directory does not exist." + "\n" +
            "User entered: " + self.value)


class ErrorAnnotFileNotExists(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorAnnotFileNotExists class.

        Parameters:
        value(str): Path to the annotation file.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the annotation file path.'''
        return ("\n\nFile does not exist." + "\n" +
            "User entered: --annotfile " + self.value)


class ErrorTableNotInDB(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorTableNotInDB class.

        Parameters:
        value(str): Name of the requested table in a sqlite database.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the name of the table.'''
        return ("\n\nThe table does not exist in the database." + "\n" +
            "Unknown table found: " + self.value)

class ErrorEntryNotInDB(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorEntryNotInDB class.

        Parameters:
        value(str): Name of the entry in a sqlite database.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the name of the entry.'''
        return ("\n\nThe entry does not exist in the database." + "\n" +
            "Unknown entry found: " + self.value)

class ErrorEntryMultipleMatches(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorEntryMultipleMatches class.

        Parameters:
        value(str): Name of the entry in a sqlite database.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the name of the entry.'''
        return ("\n\nThe entry has multiple matches in the database." + "\n" +
            "Multiple matches for entry: " + self.value)

class ErrorConfigFileNotExists(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorConfigFileNotExists class.

        Parameters:
        value(str): Path to the configuration file.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the path to the configuration file.'''
        return ("\n\nConfig file does not exist in the specified location." + "\n" +
            "Location specified: " + self.value)

class ErrorColumnsNotUnique(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorColumnsNotUnique class.

        Parameters:
        value(str): Name of the column in a sqlite database.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the name of the column.'''
        return ("\n\nThe following columns are not unique in the database:" + "\n" +
            str(self.value))

class ErrorDBNotExists(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorDBNotExists class.

        Parameters:
        value(str): Path to the database file.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the path to the configuration file.'''
        return ("\n\nDatabase does not exist." + "\n" +
            self.value)

class ErrorMaxFetchAttemptsExceeded(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorMaxFetchAttemptsExceeded class.

        Parameters:
        value(int): The number of fetch attempts.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the number of fetch attempts.'''
        return ("\n\nThe Maximum number of fetch attempts was exceeded for ID:" + "\n" +
            self.value)

class ErrorMaxReadAttemptsExceeded(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorMaxReadAttemptsExceeded class.

        Parameters:
        value(int): The number of read attempts.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the number of read attempts.'''
        return ("\n\nThe Maximum number of read attempts was exceeded for table:" + "\n" +
            self.value)

class ErrorConfigParameter(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorConfigParameter class.

        Parameters:
        value(int): The name of the config parameter.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the name of the config parameter.'''
        return ("\n\nA parameter name and/or value in the configuration file is set incorrectly:" + "\n" +
            self.value)

class ErrorConfigYAMLFormat(Exception):
    def __init__(self, value):
        '''
        The constructor for ErrorConfigYAMLFormat class.

        Parameters:
        value (str): The name of the config file.
        '''
        self.value = value
    def __str__(self):
        '''When the error is raised, print the name of the config file.'''
        return ("\n\nThe configuration file could not be loaded, please confirm that this is a proper YAML file: " + "\n" +
            self.value)



import pytest                               # Testing suite
from ncbimeta import NCBImetaUtilities      # Utility Functions
from lxml import etree                      # XML Parsing

def test_lxml(assembly_xml):
    '''Test the functionality of the lxml module'''
    column_name =  'AssemblyAccession'
    column_payload =  ['AssemblyAccession']
    column_value = ""
    root = etree.fromstring(assembly_xml, )
    tag_xpath = './/' + "Meta"
    # Step 1 going to iterate through column_payload values
    working_root = root
    column_dict = {}
    # Iterate over each path element in the column_payload list
    for i in range(0,len(column_payload)):
        # Construct an xpath recursive search query
        tag_xpath = ".//" + column_payload[i]
        # recursively search for the tag
        working_root = working_root.findall(tag_xpath)

    print(working_root)
    print(working_root[0].text)
    column_value = working_root[0].text
    column_dict[column_name] = column_value
    print(column_dict)


    #print("Findall: ", root.findall(tag_xpath))  # XPath, recursive.
    #print("Iterfind: ", root.iterfind(tag_xpath))
    #print("Findtext: ", root.findtext(tag_xpath).strip())
    #print(assembly_xml)
    assert 0

#iterfind() iterates over all Elements that match the path expression
#findall() returns a list of matching Elements
#find() efficiently returns only the first match
#findtext() returns the .text content of the first match



"""
NCBImeta Test - Annotate Replace

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                         Modules and Packages                          #
#-----------------------------------------------------------------------#

import pytest                             # Testing suite
from ncbimeta import *                    # Main Program
import os                                 # Filepath operations
import test.test_ncbimeta                 # Run the program to create test db
import subprocess                         # Execute CLI/Shell

#-----------------------------------------------------------------------#
#                           Test Function                               #
#-----------------------------------------------------------------------#

def test_annotatereplace_run():
    '''Test the NCBImetaAnnotateReplace application for run completion'''
    # Use the test database
    test_db = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test.sqlite")
    test_annotfile = os.path.join(os.path.dirname(os.path.abspath(__file__)),'test_annot.txt')
    # If the test_db doesn't alread exist, run the test cmd from test_ncbimeta
    if not os.path.exists(test_db): test_ncbimeta.test_ncbimeta_run()
    test_table = 'BioSample'
    test_cmd = ("ncbimeta/NCBImetaAnnotateReplace.py --database " + test_db +
                " --table  " + test_table +
                " --annotfile " + test_annotfile)
    # test NCBImetaAnnotateReplace through a subprocess
    returned_value = subprocess.call(test_cmd, shell=True)
    # If it returns a non-zero value, it failed
    assert returned_value == 0

#def test_annotateconcatenate_multipleEntry(tmpdir):
    #To Be done

#def test_annotatereplace_ErrorDBNotExists(tmpdir):
    #To Be done

#def test_annotatereplace_ErrorAnnotFileNotExists(tmpdir):
    #To Be done

#def test_annotatereplace_ErrorTableNotInDB(tmpdir):
    #To Be done



#!/usr/bin/env python3
"""
@author: Katherine Eaton

NCBImeta: Efficient and comprehensive metadata retrieval from the NCBI databases.
"""

# This program should only be called from the command-line
if __name__ != "__main__": quit()

#-----------------------------------------------------------------------#
#                         Modules and Packages                          #
#-----------------------------------------------------------------------#
import argparse                         # Command-line argument parsing
import sqlite3                          # Database storage and queries
import os                               # Filepath operations
import time                             # Allow sleeping of processes
import datetime                         # Get date and time for logfile
import Bio                              # Biopython NCBI API
from Bio import Entrez                  # Entrez queries (NCBI)
import yaml                             # YAML config file parsing
from lxml import etree                  # XML Parsing
import tempfile                         # Temporary file for XML parsing

from ncbimeta import NCBImetaUtilities  # NCBImeta helper functions
from ncbimeta import NCBImetaErrors     # NCBImeta Error classes



#-----------------------------------------------------------------------#
#                            Argument Parsing                           #
#-----------------------------------------------------------------------#

parser = argparse.ArgumentParser(description='NCBImeta: Efficient and comprehensive metadata retrieval from the NCBI databases.',
                                 add_help=True)

# Argument groups for the program
mandatory_parser = parser.add_argument_group('mandatory')

parser.add_argument('--config',
                    help = 'Path to the yaml configuration file (ex. config.yaml).',
                    action = 'store',
                    dest = 'configPath',
                    required = True)

parser.add_argument('--flat',
                    help = 'Do not create organization directories, output all files to output directory.',
                    action = 'store_true',
                    dest = 'flatMode')

parser.add_argument('--version',
                    action='version',
                    version='%(prog)s v0.4.2')

# Retrieve user parameters
args = vars(parser.parse_args())

config_path = args['configPath']
flat_mode = args['flatMode']

#------------------------------------------------------------------------------#
#                            Error Catching                                    #
#------------------------------------------------------------------------------#

# Check if configuration file exists
if not os.path.exists(config_path):
    raise NCBImetaErrors.ErrorConfigFileNotExists(config_path)

# Load the YAML configuration file
with open(config_path) as config_file:
    config_data = yaml.load(config_file, Loader=yaml.FullLoader)
    if config_data is None:
        raise NCBImetaErrors.ErrorConfigYAMLFormat(config_file)

# Retrieve configuration file values and error catching
#--- Output Directory ---#
try:
    CONFIG_OUTPUT_DIR = config_data["OUTPUT_DIR"]
except KeyError:
    raise NCBImetaErrors.ErrorConfigParameter("OUTPUT_DIR")
#--- User Email ---#
try:
    CONFIG_EMAIL = config_data["EMAIL"]
except KeyError:
    raise NCBImetaErrors.ErrorConfigParameter("EMAIL")
#--- User API Key ---#
try:
    CONFIG_API_KEY = config_data["API_KEY"]
except KeyError:
    raise NCBImetaErrors.ErrorConfigParameter("API_KEY")
#--- Force pausing in between record fetching ---#
try:
    CONFIG_FORCE_PAUSE_SECONDS = config_data["FORCE_PAUSE_SECONDS"]
except KeyError:
    raise NCBImetaErrors.ErrorConfigParameter("FORCE_PAUSE_SECONDS")
#--- Database file name---#
try:
    CONFIG_DATABASE = config_data["DATABASE"]
except KeyError:
    raise NCBImetaErrors.ErrorConfigParameter("DATABASE")
#--- NCBI Tables to Search (list) ---#
try:
    CONFIG_TABLES = config_data["TABLES"]
except KeyError:
    raise NCBImetaErrors.ErrorConfigParameter("TABLES")
#--- Search terms for each table (dict) ---#
try:
    CONFIG_SEARCH_TERMS = config_data["SEARCH_TERMS"]
except KeyError:
    raise NCBImetaErrors.ErrorConfigParameter("SEARCH_TERMS")
#--- Table Columns/Metadata to retrieve (list of dict)---#
try:
    CONFIG_TABLE_COLUMNS = config_data["TABLE_COLUMNS"]
except KeyError:
    raise NCBImetaErrors.ErrorConfigParameter("TABLE_COLUMNS")

#--- Print the retrieved config file parameters ---#
print(
"\n" + "NCBImeta was run with the following options: " + "\n" +
"\t" + "Config File: " + str(config_path) + "\n" +
"\t" + "Output Directory: " + str(CONFIG_OUTPUT_DIR) + "\n" +
"\t" + "Email: " + str(CONFIG_EMAIL) + "\n" +
"\t" + "User Database: " + str(CONFIG_DATABASE) + "\n" +
"\t" + "Tables: " + str(CONFIG_TABLES) + "\n" +
"\t" + "Search Terms: ", flush = True)
for table_search_term in CONFIG_SEARCH_TERMS:
    print("\t\t" + str(table_search_term), flush = True)
print("\n", flush = True)

# Check if output dir exists
if not os.path.exists(CONFIG_OUTPUT_DIR):
    raise NCBImetaErrors.ErrorOutputDirNotExists(CONFIG_OUTPUT_DIR)

# Flat mode checking
if flat_mode:
    print("Flat mode was requested, organizational directories will not be used.", flush = True)
    DB_DIR = os.path.join(CONFIG_OUTPUT_DIR, "")
    LOG_PATH = CONFIG_OUTPUT_DIR

# Or Create accessory directory (ex. log, data, database, etc.)
elif not flat_mode:
    print("Flat mode was not requested, organization directories will be used.", flush = True)
    NCBImetaUtilities.check_accessory_dir(CONFIG_OUTPUT_DIR)
    DB_DIR = os.path.join(CONFIG_OUTPUT_DIR, "", "database", "")
    LOG_PATH = os.path.join(CONFIG_OUTPUT_DIR, "", "log")

# Database path (depending on flat mode or not)
DB_PATH = os.path.join(DB_DIR, "", CONFIG_DATABASE)

#------------------------------------------------------------------------------#
#                       Database connection                                    #
#------------------------------------------------------------------------------#
if not os.path.exists(DB_PATH):
    print("\n" + "Creating database: " + DB_PATH, flush = True)
    conn = sqlite3.connect(DB_PATH)
    conn.commit()
    print("\n" + "Connected to database: " + DB_PATH, flush = True)

elif os.path.exists(DB_PATH):
    conn = sqlite3.connect(DB_PATH)
    conn.commit()
    print("\n" + "Connected to database: " + DB_PATH, flush = True)

#-----------------------------------------------------------------------#
#                      CONSTANTS and Config                             #
#-----------------------------------------------------------------------#

XPATH_SPECIAL_CHAR =['~', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')',
                     '+', '{', '}', '|', ':', '"', '<', '>', '?', '`', '=',
                     '[', ']', '\\', ';', '‘', ',', '.', '/']

DB_VALUE_SEP = ";"

# lxml parser
LXML_CDATA_PARSER = etree.XMLParser(strip_cdata=False)

#------------------------------------------------------------------------------#
#                       Database Processing Function                           #
#------------------------------------------------------------------------------#

def UpdateDB(table, output_dir, database, email, search_term, table_columns, log_path, db_dir, api_key, force_pause_seconds):
    '''
    Update the contents of a local sqlite database using records retrieved from NCBI as configured by the user.

    Parameters:
    table (str): Name of the NCBI database to search.
    output_dir (str): Path to the directory where output is written.
    database (str): Filename of the local sqlite database.
    email (str): User email.
    search_term (str): Entrez search query.
    table_columns(dict): Dictionary of column name and API name as value, ex. {AssemblyGenbankID : GbUid}.
    log_path(str): Path to the directory where the logfile is stored in.
    db_dir(str): Path to the directory where the database is stored in.
    api_key(str): NCBI user account API Key.
    force_pause_seconds(float): Number of seconds to wait in between fetch read_attempts.
    '''

    print("\nCreating/Updating the " + table + " table using the following parameters: " + "\n" +
    "\t" + "Database: " + "\t\t" + database + "\n" +
    "\t" + "Search Term:" + "\t" + "\t" + search_term + "\n" +
    "\t" + "Email: " + "\t\t\t" + email + "\n" +
    "\t" + "API Key: " + "\t\t\t" + api_key + "\n" +
    "\t" + "Output Directory: " + "\t" + output_dir + "\n\n", flush = True)


    Entrez.email = email
    Entrez.api_key = api_key
    # Allow a maximum of 3 tries for error catching before exiting program
    Entrez.max_tries = 3
    # Sleep for 1 second after an error has been generated before retrying
    Entrez.sleep_between_tries = 1

    #---------------------------------------------------------------------------#
    #                                File Setup                                 #
    #---------------------------------------------------------------------------#
    # Name of Log File
    log_file_path = os.path.join(LOG_PATH, "",
                                os.path.splitext(database)[0] + "_" + table + ".log")

    # Check if the file already exists, either write or append to it.
    if os.path.exists(log_file_path):
        log_file = open(log_file_path, "a")
    else:
        log_file = open(log_file_path, "w")

    #--------------------------------------------------------------------------#
    #                                SQL Setup                                 #
    #--------------------------------------------------------------------------#

    # Connect to database and establish cursor for commands.
    conn = sqlite3.connect(os.path.join(db_dir, "", database))
    cur = conn.cursor()

    ## Create the database, with dynamic variables from config file
    sql_query = ("Create TABLE IF NOT EXISTS " + table +
    " (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, " +
    table + "_id TEXT")

    for column_name_dict in table_columns:
        column_name = list(column_name_dict.keys())[0]
        # By default, every user-specified column is type TEXT
        sql_query += ", " + column_name + " TEXT"
    sql_query += ")"

    cur.execute(sql_query)

    #-----------------------------------------------------------------------#
    #                          Entrez Search                                #
    #-----------------------------------------------------------------------#
    # Read the record, check for http, url, and runtime errors
    read_succeed = False
    read_attempts = 0

    # Database reading and entrez searching occur in a while loop to catch errors
    while not read_succeed and read_attempts < Entrez.max_tries:
        kwargs = {"db": table.lower(), "term":search_term, "retmax":"9999999"}
        entrez_method = Entrez.esearch
        # Possible urllib error and RuntimeErrors occurring in the next line
        handle = NCBImetaUtilities.HTTPErrorCatch(entrez_method, Entrez.max_tries, Entrez.sleep_between_tries, **kwargs)
        try:
            record = Entrez.read(handle)
            read_succeed = True
        except RuntimeError:
            read_attempts += 1
            print("Runtime Error encountered. Sleeping for " + str(Entrez.sleep_between_tries) + " seconds before retrying.")
            time.sleep(Entrez.sleep_between_tries)

    if read_attempts == Entrez.max_tries and not read_succeed:
        raise ErrorMaxReadAttemptsExceeded(table)

    # Count total number of entries, create counter
    num_records = int(record['Count'])
    num_processed = 0

    #-----------------------------------------------------------------------#
    #                          Iterate Through ID List                      #
    #-----------------------------------------------------------------------#

    for ID in record['IdList']:
        #-------------------Progress Log and Entry Counter-------------------#
        # Increment entry counter and record progress to screen
        num_processed += 1
        print("ID: " + ID, flush = True)
        print("Processing record: " +
               str(num_processed) + \
               "/" + str(num_records), flush = True)

        #------------Check if Record Already Exists in Database------------#
        sql_query = ("SELECT EXISTS(SELECT " + table + "_id FROM " +
                    table + " WHERE " + table + "_id=?)")
        cur.execute(sql_query, (ID,))

        # 0 if not found, 1 if found
        record_exists = cur.fetchone()[0]

        # If the record_exists, skip the whole next part (ie. "continue" to next record)
        if record_exists:
            continue
        '''
        IMPORTANT:
        The ID should not exists in the table UNLESS the record was fully parsed.
        ie. The database does not get updated until the end of each record.
        '''
        # This is the sleep command before implementing the HTTPerror catching in next section
        # This is controlled by the user configuration file
        time.sleep(force_pause_seconds)

        #---------------If the table isn't in Database, Add it------------#
        # The Assembly table cannot be retrieved using efetch, only docsum esummary
        if table.lower() == "assembly":
            # Use the http function to return a record summary, but wrapped in HTTP error checking
            kwargs = {"db":table.lower(), "id":ID, "retmode": "xml"}
            entrez_method = Entrez.esummary
        else:
            # We're working with any other table instead, use efetch and get xml
            kwargs = {"db": table.lower(), "id": ID, "retmode": "xml"}
            if table.lower() == "nucleotide":
                kwargs["rettype"] = "gb"
            entrez_method = Entrez.efetch

        # ID_handle is an _io.TextIOWrapped object, which originally had utf-8 encoding
        ID_handle = NCBImetaUtilities.HTTPErrorCatch(entrez_method, Entrez.max_tries, Entrez.sleep_between_tries, **kwargs)

        # Ideal world: Pass an undecoded string to the xml parser
        # Could be accomplished by opening in binary ('rb')
        # tempfiles by default are opened as mode='w+b'
        with tempfile.NamedTemporaryFile(delete=False) as temp_b:
            # Write the data from ID_handle to a temporary file (binary)
            for line in ID_handle: temp_b.write(str.encode(line))
            temp_b.close()
            # Read the data as binary, into the XML parser. Avoids encoding issues
            with open(temp_b.name, 'rb') as xml_source:
                ID_root = etree.parse(xml_source, parser=LXML_CDATA_PARSER)

        #----------------------------------------------------------------------#
        #                         NCBI Record Parsing                          #
        #----------------------------------------------------------------------#

        column_dict = {}
        # Add ID to the dictionary
        column_dict[table + "_id"] = [ID]
        # A special dictionary for gbseq annotations
        gbseq_dict = {}
        # Iterate through each column to search for metadata
        for column in table_columns:
            column_name = list(column.keys())[0]
            column_value = []
            column_payload = list(column.values())[0]
            column_payload = column_payload.split(", ")
            # Initialize with empty values
            column_dict[column_name] = column_value

            #-------------------------------------------------------#
            #   XML Parse for node or attribute
            #-------------------------------------------------------#
            working_root =ID_root
            # If there are special character, this query should not be used for xpath!!
            bool_special_char = False
            for char in XPATH_SPECIAL_CHAR:
                for xquery in column_payload:
                    if char in xquery:
                        bool_special_char = True
            # If no special characters, run xpath search Functions
            if not bool_special_char:
                NCBImetaUtilities.xml_search(working_root, column_payload, column_payload[0], column_name, column_dict)

            # Special parsing for GBSeq_comment
            # If we're on the GBSeq_comment element and the comment was added to the dictionary
            if "GBSeq_comment" in column_payload and len(column_dict[column_name]) > 0:
                comment = column_dict[column_name][0]
                # Fix the CDS vs CDSs ambiguity
                comment = comment.replace("CDSs", "CDS")
                # comment is initialize subdivided by semi-colons
                split_comment = comment.split(";")
                for item in split_comment:
                    # Further subdivided by double colons
                    split_item = item.split("::")
                    # The elements we're interested have the :: otherwise skip
                    if len(split_item) < 2: continue
                    # Left side is the column name, right side is the metadata
                    split_key = split_item[0].lstrip(" ").rstrip(" ")
                    split_value = split_item[1].lstrip(" ").rstrip(" ")
                    gbseq_dict[split_key] = split_value

            # If the value was still empty, check for gbseq comment
            if column_payload[0] in gbseq_dict:
                column_dict[column_name].append(gbseq_dict[column_payload[0]])

        # Add quotations around each value for sql insertion
        for key in column_dict:
            # Remove empty string elements
            while "" in column_dict[key]: column_dict[key].remove("")
            # Remove quotations from each list element
            for i in range(0,len(column_dict[key])):
                column_dict[key][i] = column_dict[key][i].replace("\"","")
            # The following is to help with single quotes inside
            column_dict[key] = "\"" + DB_VALUE_SEP.join(column_dict[key]) + "\""

        # Write the column values to the db with dynamic variables
        sql_dynamic_table = "INSERT INTO " + table + " ("
        sql_dynamic_vars = ",".join([column for column in column_dict.keys()]) + ") "
        sql_dynamic_qmarks = "VALUES (" + ",".join(["?" for column in column_dict.keys()]) + ") "
        sql_dynamic_values = " VALUES (" + ",".join([column_dict[column] for column in column_dict.keys()]) + ")"
        sql_query = sql_dynamic_table + sql_dynamic_vars + sql_dynamic_values
        sql_query_q = sql_dynamic_table + sql_dynamic_vars + sql_dynamic_qmarks
        cur.execute(sql_query)

        # Write to logfile
        now = datetime.datetime.now()
        log_file.write("[" + str(now) + "]" +
                     "\t" + "New entry added with ID:" +
                     "\t" + ID + "." + "\n")
        conn.commit()


    # CLEANUP
    conn.commit()
    cur.close()
    log_file.close()

#------------------------------------------------------------------------------#
#                     Iterate Through Tables in Config File                    #
#------------------------------------------------------------------------------#
for table in CONFIG_TABLES:
    OUTPUT_DIR = CONFIG_OUTPUT_DIR
    DATABASE = CONFIG_DATABASE
    EMAIL = CONFIG_EMAIL
    # Since API can be empty, force it to empty string for printing
    API_KEY = CONFIG_API_KEY
    if not API_KEY: API_KEY = ""
    FORCE_PAUSE_SECONDS = CONFIG_FORCE_PAUSE_SECONDS
    # Config search terms, match table to table name
    for index,search_term_table in enumerate(CONFIG_SEARCH_TERMS):
        table_name = list(search_term_table)[0]
        if table_name == table:
            break
    SEARCH_TERM = CONFIG_SEARCH_TERMS[index][table]
    # Table columns, match table to table name
    for index,table_column_dict in enumerate(CONFIG_TABLE_COLUMNS):
        table_name = list(table_column_dict)[0]
        if table_name == table:
            break
    TABLE_COLUMNS = CONFIG_TABLE_COLUMNS[index][table]

    # Check for duplicate column names
    # -- Note this is only PER TABLE (Checking for duplicate between table only happens
    # -- later in the database Join script)
    table_col_names = []
    for element in TABLE_COLUMNS: table_col_names += list(element.keys())
    dupl_table_col_names = set([col for col in table_col_names if table_col_names.count(col) > 1])
    if len(dupl_table_col_names) > 0:
        raise NCBImetaErrors.ErrorColumnsNotUnique(dupl_table_col_names)

    # Call the main database updating function
    UpdateDB(table, OUTPUT_DIR, DATABASE, EMAIL, SEARCH_TERM, TABLE_COLUMNS, LOG_PATH, DB_DIR, API_KEY, FORCE_PAUSE_SECONDS)



#!/usr/bin/env python3
"""
NCBImeta Join Tool - Joins accessory tables to an anchor table.

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                         Modules and Packages                          #
#-----------------------------------------------------------------------#

import argparse                         # Command-line argument parsing
import sqlite3                          # Database storage and queries
import os                               # Filepath operations
import datetime                         # Get date and time for logfile

from ncbimeta import NCBImetaUtilities  # NCBImeta helper functions
from ncbimeta import NCBImetaErrors     # NCBImeta Error classes

#-----------------------------------------------------------------------#
#                            Argument Parsing                           #
#-----------------------------------------------------------------------#

parser = argparse.ArgumentParser(description=("NCBImeta Join Tool - Joins accessory tables to an anchor table."),
                                 add_help=True)

mandatory = parser.add_argument_group('mandatory')
bonus = parser.add_argument_group('bonus')

mandatory.add_argument('--database',
                    help='Path to the sqlite database generated by NCBImeta.',
                    type = str,
                    action = 'store',
                    dest = 'dbName',
                    required=True)

mandatory.add_argument('--anchor',
                    help='Table in NCBImeta database to use as an anchor for joining',
                    type = str,
                    action = 'store',
                    dest = 'dbAnchor',
                    required=True)

mandatory.add_argument('--accessory',
                    help='Accessory tables to join to the anchor as space separated str list ex. "BioProject Assembly" ',
                    type = str,
                    action = 'store',
                    dest = 'dbAccessory',
                    required=True)

mandatory.add_argument('--final',
                    help="Name of the final table after the join",
                    type = str,
                    action = 'store',
                    dest = 'dbFinal',
                    required=True)

mandatory.add_argument('--unique',
                    help='List of unique values in anchor table, that could be found in all accessory tables to be joined ex. "Accession BioProject"',
                    type = str,
                    action = 'store',
                    dest = 'dbUnique',
                    required=True)

parser.add_argument('--version',
                    action='version',
                    version='%(prog)s v0.4.2')


args = vars(parser.parse_args())

db_name = args['dbName']
db_anchor = args['dbAnchor']
db_accessory_str = args['dbAccessory']
db_accessory_list = db_accessory_str.split(" ")
db_final = args['dbFinal']
db_all_tables = [db_anchor] + db_accessory_list
unique_header_str = args['dbUnique']
unique_header_list = unique_header_str.split(" ")
DB_VALUE_SEP = ";"

#-----------------------------------------------------------------------#
#                           Argument Checking                           #
#-----------------------------------------------------------------------#


#---------------------------Check Database------------------------------#

if os.path.exists(db_name):
    conn = sqlite3.connect(db_name)
    print('\nOpening database: ' + db_name + "\n", flush = True)
else:
    raise NCBImetaErrors.ErrorDBNotExists(db_name)

# no errors were raised, safe to connect to db
cur = conn.cursor()

#---------------------------Check Tables---------------------------------#

if not NCBImetaUtilities.table_exists(cur, db_anchor):
    raise NCBImetaErrors.ErrorTableNotInDB(db_anchor)
for table in db_accessory_list:
    if not NCBImetaUtilities.table_exists(cur, table):
        raise NCBImetaErrors.ErrorTableNotInDB(table)

#-----------------------------------------------------------------------#
#                                File Setup                             #
#-----------------------------------------------------------------------#

# get list of column names in anchor table
cur.execute('''SELECT * FROM {}'''.format(db_anchor))
db_col_names = [description[0] for description in cur.description if description[0] != "id"]
anchor_col_names = [description[0] for description in cur.description]

# Check to make sure the unique header is present in the anchor table
for unique_header in unique_header_list:
    if unique_header not in db_col_names:
        print("Column not in DB: " + unique_header + ".", flush = True)
        raise NCBImetaErrors.ErrorEntryNotInDB(unique_header)

# get list of column names in accessory tables
for table in db_accessory_list:
   cur.execute('''SELECT * FROM {}'''.format(table))
   for header in [description[0] for description in cur.description if description[0] != "id"]:
       db_col_names.append(header)

# Prepare the columns for the final join table
dupl_col_names = set([col for col in db_col_names if db_col_names.count(col) > 1])
if len(dupl_col_names) > 0:
    raise NCBImetaErrors.ErrorColumnsNotUnique(dupl_col_names)

#-----------------------------------------------------------------------#
#                              Init Join Table                          #
#-----------------------------------------------------------------------#

## Create the database with dynamic variables to store joined info
sql_query = ("Create TABLE IF NOT EXISTS " + db_final +
        " (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE ")

for column_name in db_col_names:
    sql_query += ", " + column_name + " TEXT"
sql_query += ")"
cur.execute(sql_query)

#-----------------------------------------------------------------------#
#                              Join Tables                              #
#-----------------------------------------------------------------------#

# Retrieve all the anchor table records
cur.execute('''SELECT * FROM {}'''.format(db_anchor))
fetch_records = cur.fetchall()

num_processed = 0
num_records = len(fetch_records)


# Iterate through each record
for record in fetch_records:
    # Initialize/Reinitialize the master table value dictionary
    master_column_dict = {}
    for col in db_col_names: master_column_dict[col] = "NULL"

    # Store the anchor table values
    for i in range(0,len(record)):
        # Skip id
        if anchor_col_names[i] == "id": continue
        if record[i] is None:
            record_val = "NULL"
        else:
            record_val = "'" + str(record[i]) + "'"
        master_column_dict[anchor_col_names[i]] = record_val

    # Find the first unique column and its headers
    unique_val = master_column_dict[unique_header_list[0]]

    #-------------------Progress Log and Entry Counter-------------------#
    # Increment entry counter and record progress to screen

    num_processed += 1
    print("Unique Value: " + unique_val, flush = True)
    print("Processing record: " +
      str(num_processed) + \
     "/" + str(num_records), flush = True)


    # Check if this record already exists in the master join table
    sql_query = ("SELECT EXISTS(SELECT " + unique_header_list[0] + " FROM " +
                 db_final + " WHERE " + unique_header_list[0] + "=" +
                 unique_val + ")" )
    cur.execute(sql_query)

    # 0 if not found, 1 if found
    record_exists = cur.fetchone()[0]
    if record_exists: continue

    # Now grab the associated records in the accessory tables
    sql_query = '''SELECT {0} FROM {1} WHERE {2}={3}'''.format(",".join(unique_header_list),
                                                               db_anchor,
                                                               unique_header_list[0],
                                                               unique_val)
    cur.execute(sql_query)
    fetch_unique_vals_anchor = cur.fetchall()


    for unique_values in fetch_unique_vals_anchor:
        # Convert from tuple to list for index positions
        unique_values = list(unique_values)
        # Ignore any supposedly unique values that wound up being missing
        if None in unique_values: unique_values.remove(None)

        # search for this value in each accessory table
        for table in db_accessory_list:
            # Get list of each column
            cur.execute(''' SELECT * FROM {}'''.format(table))
            table_col_names = [description[0] for description in cur.description]

            # Initialize matching variables
            match_found=False
            match_column = ""
            # Iterate through each possible unique values (or until match is found)
            for uniq_val in unique_values:
                # Deal with unicode/str
                if type(uniq_val) == int: val=str(uniq_val)
                elif type(uniq_val) == str: uniq_val=uniq_val.encode('utf-8')

                # Iterate through each column
                for table_col in table_col_names:
                    cur.execute('''SELECT {0} FROM {1}'''.format(table_col, table))
                    table_col_vals = cur.fetchall()
                    # Search through every value for a match
                    for val in table_col_vals:
                        if type(val[0]) == int: val=str(val[0])
                        elif type(val[0]) == str: val=val[0].encode('utf-8')

                        # If it's a match, store the value, and set the boolean flag
                        if val == uniq_val:
                            match_found=True
                            match_column=table_col
                            match_val = val
                            # Found the match, stop searching through vals in this column
                            break

                    # Found a match, stop searching through columns
                    if match_found: break
                # Found  a match, stop searching values
                if match_found: break

            if match_found:
                query=('''SELECT * FROM {0} WHERE {1}={2}'''.format(table,
                                                                match_column,
                                                                "'" + match_val.decode('utf-8') + "'"))
                cur.execute(query)
                match_records = cur.fetchall()
                record_dict = {}

                # If multiple records are found, concatenate into a pseudo single record
                if len(match_records) > 1:
                    match_records_concat = [None] * len(match_records[0])
                    for i in range(0,len(match_records[0])):
                        tmp_record_list = []
                        for record in match_records:
                            if record[i] is None or record[i] == "": continue
                            elif type(record[i]) == int: tmp_record_list.append(str(record[i]))
                            else:
                                tmp_record_list.append(record[i])
                        if len(tmp_record_list) > 1:
                            # Check if all values are identical
                            dupl_values = set([val for val in tmp_record_list if tmp_record_list.count(val) == len(tmp_record_list)])
                            if len(dupl_values) == 1:
                                match_records_concat[i] = list(dupl_values)[0]
                            else: match_records_concat[i] = DB_VALUE_SEP.join(tmp_record_list)

                        else: match_records_concat[i] = None
                    match_records = [match_records_concat]


                match_records = list(match_records[0])
                for i in range(0,len(table_col_names)):
                    # Check if this is a valid column name in the master join table
                    if table_col_names[i] not in master_column_dict.keys(): continue
                    record_val = match_records[i]
                    # Check for None and handle unicode
                    if record_val is None: record_val = "NULL"
                    # FIX THIS UNICODE MESS
                    else:
                        try:
                            record_val = "'" + str(record_val) + "'"
                        except:
                            record_val = "'" + record_val.encode('utf-8') + "'"
                    # Assign record to dictionary
                    master_column_dict[table_col_names[i]] = record_val
                    master_column_dict[table_col_names[i]] = "\"" + DB_VALUE_SEP.join(master_column_dict[table_col_names[i]]) + "\""


        # We've now finished processing all accessory tables
        # Add values to new join table
        sql_dynamic_table = "INSERT INTO " + db_final + " ("
        sql_dynamic_vars = ",".join([column for column in master_column_dict.keys()]) + ")"
        sql_dynamic_values = " VALUES (" + ",".join([master_column_dict[column] for column in master_column_dict.keys()]) + ")"
        sql_query = sql_dynamic_table + sql_dynamic_vars + sql_dynamic_values
        cur.execute(sql_query)
        # Save Changes
        conn.commit()

#-----------------------------------------------------------------------#
#                                    Cleanup                            #
#-----------------------------------------------------------------------#
# Commit changes
conn.commit()
print("Closing database: " + db_name, flush = True)
cur.close()



"""
NCBImeta Test - Utility Functions

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                         Modules and Packages                          #
#-----------------------------------------------------------------------#

import pytest                               # Testing suite
from ncbimeta import NCBImetaUtilities      # Utility Functions
import os                                   # Filepath and directory operations
import sqlite3                              # Database storage and queries
from lxml import etree                      # XML Parsing
from Bio import Entrez                      # Entrez queries (NCBI)
#-----------------------------------------------------------------------#
#                           Test Function                               #
#-----------------------------------------------------------------------#

def test_check_accessory_dir(tmpdir):
    '''Test the utility function check_accessory_dir (create log/ and database/).'''
    tmpdir = tmpdir.strpath
    NCBImetaUtilities.check_accessory_dir(tmpdir)
    assert os.path.exists(os.path.join(tmpdir,"log")) and os.path.exists(os.path.join(tmpdir,"database"))

def test_table_exists(tmpdir):
    '''Test the utility function table_exists (check if Table is present in sqlite db)'''
    # Connect to database and establish cursor for commands.
    tmpdir = tmpdir.strpath
    test_db = os.path.join(tmpdir, "test.sqlite")
    conn = sqlite3.connect(test_db)
    cur = conn.cursor()

    ## Create the database with a test Table
    table_name = "TestTable"
    sql_query = "Create TABLE IF NOT EXISTS " + table_name + " (id INTEGER)"
    cur.execute(sql_query)

    # Test Function Call
    assert NCBImetaUtilities.table_exists(cur, table_name)

def test_xml_search_nodetext():
    '''Test the utility function xml_search (retrieve xml text node)'''
    test_xml = "<Root><Assembly><AssemblyAccession>GCA_003086155.1</AssemblyAccession></Assembly></Root>"
    test_root = etree.fromstring(test_xml)
    test_search_list = ['Assembly','AssemblyAccession']
    test_current_tag = test_search_list[0]
    test_column_name = 'AssemblyAccession'
    test_xml_dict = {}
    test_xml_dict[test_column_name] = []
    expect_xml_dict = {test_column_name : ['GCA_003086155.1'] }
    NCBImetaUtilities.xml_search(test_root, test_search_list, test_current_tag , test_column_name, test_xml_dict)
    #assert 0
    assert test_xml_dict == expect_xml_dict

def test_xml_search_nodeattr():
    '''Test the utility function xml_search (retrieve xml attr text)'''
    test_xml = "<Root><Experiment acc='SRX6977650'/></Root>"
    test_root = etree.fromstring(test_xml)
    test_search_list = ['Experiment', 'acc']
    test_current_tag = test_search_list[0]
    test_column_name = 'ExperimentAccession'
    test_xml_dict = {}
    test_xml_dict[test_column_name] = []
    expect_xml_dict = {test_column_name : ['SRX6977650'] }
    NCBImetaUtilities.xml_search(test_root, test_search_list, test_current_tag , test_column_name, test_xml_dict)

    assert test_xml_dict == expect_xml_dict

def test_xml_search_nodeattrtext():
    '''Test the utility function xml_search (retrieve xml attr text)'''
    #test_xml = "<Root><LIBRARY_LAYOUT><PAIRED/></LIBRARY_LAYOUT></Root>"
    test_xml = "<Root><Stats><Stat category='chromosome_count' sequence_tag='all'>1</Stat></Stats></Root>"
    test_root = etree.fromstring(test_xml)
    test_search_list = ['Stat', 'category', 'chromosome_count']
    test_current_tag = test_search_list[0]
    test_column_name = 'ChromosomeCount'
    test_xml_dict = {}
    test_xml_dict[test_column_name] = []
    expect_xml_dict = {test_column_name : ['1'] }
    NCBImetaUtilities.xml_search(test_root, test_search_list, test_current_tag , test_column_name, test_xml_dict)
    assert test_xml_dict == expect_xml_dict

def test_xml_search_nodechildnode():
    '''Test the utility function xml_search (retrieve xml attr text)'''
    test_xml = "<Root><LIBRARY_LAYOUT><PAIRED/></LIBRARY_LAYOUT></Root>"
    test_root = etree.fromstring(test_xml)
    test_search_list = ['LIBRARY_LAYOUT']
    test_current_tag = test_search_list[0]
    test_column_name = 'LibraryLayout'
    test_xml_dict = {}
    test_xml_dict[test_column_name] = []
    expect_xml_dict = {test_column_name : ['PAIRED'] }
    NCBImetaUtilities.xml_search(test_root, test_search_list, test_current_tag , test_column_name, test_xml_dict)
    assert test_xml_dict == expect_xml_dict

def test_HTTPErrorCatch(tmpdir):
    '''Test the utility function HTTPErrorCatch (catch HTTP Errors)'''
    # Test query assembly
    test_ID = '5025191'
    test_email = 'ktmeaton@gmail.com'
    Entrez.email = test_email
    test_table = 'Assembly'
    test_max_tries = 10
    test_total_tries = 10
    test_sleep_between_tries = 0
    test_kwargs = {"db":test_table.lower(), "id":test_ID}
    test_entrez_method = Entrez.esummary

    for i in range(1, test_total_tries):
        ID_handle = NCBImetaUtilities.HTTPErrorCatch(test_entrez_method, test_max_tries, test_sleep_between_tries, **test_kwargs)
    try:
        ID_record = Entrez.read(ID_handle, validate=False)
        assert 1
    except RuntimeError:
        # The ID_handle was not succesfully retrieved or the data is not correct
        assert 0



#!/usr/bin/env python3
"""
NCBImeta Export Tool - Export SQLite database to tabular text files.

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                            Argument Parsing                           #
#-----------------------------------------------------------------------#

import argparse                         # Command-line argument parsing
import sqlite3                          # Database storage and queries
import os                               # Filepath operations

from ncbimeta import NCBImetaErrors     # NCBImeta Error classes

#-----------------------------------------------------------------------#
#                            Argument Parsing                           #
#-----------------------------------------------------------------------#

parser = argparse.ArgumentParser(description=("NCBImeta Export Tool - Export SQLite database to tabular text files."),
                                 add_help=True)

mandatory = parser.add_argument_group('mandatory')

mandatory.add_argument('--database',
                    help='Path to the sqlite database generated by NCBImeta.',
                    type = str,
                    action = 'store',
                    dest = 'dbName',
                    required=True)

mandatory.add_argument('--outputdir',
                    help = 'Output directory.',
                    action = 'store',
                    dest = 'outputDir',
                    required = True)

parser.add_argument('--version',
                    action='version',
                    version='%(prog)s v0.4.2')


args = vars(parser.parse_args())

db_name = args['dbName']
output_dir = args['outputDir']



#-----------------------------------------------------------------------#
#                           Argument Checking                           #
#-----------------------------------------------------------------------#

# Check if database exists
if os.path.exists(db_name):
    conn = sqlite3.connect(db_name)
    print('\nOpening database: ' + db_name, flush = True)
else:
    raise NCBImetaErrors.ErrorDBNotExists(db_name)

# Check if output dir exists
if not os.path.exists(output_dir):
    raise NCBImetaErrors.ErrorOutputDirNotExists(output_dir)

# no errors were raised, safe to connect to db
cur = conn.cursor()

#-----------------------------------------------------------------------#
#                         Process Database                              #
#-----------------------------------------------------------------------#

# Get a list of tables
cur.execute("SELECT name FROM sqlite_master WHERE type='table';")
table_list = cur.fetchall()

# Iterate through each table
for table in table_list:
    # Skip sqlite_sequence default table
    if table[0] == "sqlite_sequence": continue
    # Convert from tuple where second value is empty
    table_name=table[0]

    # Name of output table file
    db_basename = os.path.basename(db_name).split(".")[0]
    table_file_path = os.path.join(output_dir, "", db_basename + "_" + table_name + ".txt")
    table_file = open(table_file_path, "w")

    # get list of column names in Table
    query = "SELECT * FROM {0}".format(table_name)
    cur.execute(query)
    table_col_names = [description[0] for description in cur.description]

    # Retrieve and write the header
    header = ""
    for column in table_col_names:
        header += str(column) + "\t"
    header = header.rstrip("\t") + "\n"
    table_file.write(header)

    # Retrieve and write the lines
    query = "SELECT * FROM {0}".format(table_name)
    for row in cur.execute(query):
        line = ""
        for cell in row:
            line += str(cell) + "\t"
        line = line.rstrip("\t")  + "\n"
        table_file.write(line)

    # Close table file
    table_file.close()

#-----------------------------------------------------------------------#
#                                    Cleanup                            #
#-----------------------------------------------------------------------#

print("Closing database: " + db_name, flush = True)
cur.close()



#!/usr/bin/env python3
"""
NCBImeta Annotation Tool - Replaces database fields with values in curated annotation file.

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                            Argument Parsing                           #
#-----------------------------------------------------------------------#

import argparse                         # Command-line argument parsing
import sqlite3                          # Database storage and queries
import datetime                         # Get date and time for logfile
import os                               # Filepath operations

from ncbimeta import NCBImetaErrors     # NCBImeta Error Classes
from ncbimeta.NCBImetaUtilities import table_exists

#-----------------------------------------------------------------------#
#                            Argument Parsing                           #
#-----------------------------------------------------------------------#

parser = argparse.ArgumentParser(description=("NCBImeta Annotation Tool - Replaces database fields with values in curated annotation file."),
                                 add_help=True)

mandatory = parser.add_argument_group('mandatory')
bonus = parser.add_argument_group('bonus')

mandatory.add_argument('--database',
                    help='Path to the sqlite database generated by NCBImeta.',
                    type = str,
                    action = 'store',
                    dest = 'dbName',
                    required=True)

mandatory.add_argument('--table',
                    help='Table in NCBImeta database to modify',
                    type = str,
                    action = 'store',
                    dest = 'dbTable',
                    required=True)

mandatory.add_argument('--annotfile',
                    help='Path to annotation file. The first column must contain a field that is unique to the record (ex. Accession)',
                    type = str,
                    action = 'store',
                    dest = 'annotFile',
                    required=True)

parser.add_argument('--version',
                    action='version',
                    version='%(prog)s v0.4.2')


args = vars(parser.parse_args())

db_name = args['dbName']
db_table = args['dbTable']
annot_file_name = args['annotFile']


#-----------------------------------------------------------------------#
#                           Argument Checking                           #
#-----------------------------------------------------------------------#


#---------------------------Check Database------------------------------#

if os.path.exists(db_name):
    conn = sqlite3.connect(db_name)
    print('\nOpening database: ' + db_name, flush = True)
else:
    raise NCBImetaErrors.ErrorDBNotExists(db_name)

if not os.path.exists(annot_file_name):
    raise NCBImetaErrors.ErrorAnnotFileNotExists(annot_file_name)

# no errors were raised, safe to connect to db
cur = conn.cursor()

#---------------------------Check Table---------------------------------#

if not table_exists(cur, db_table):
    raise NCBImetaErrors.ErrorTableNotInDB(db_table)


#-----------------------------------------------------------------------#
#                                File Setup                             #
#-----------------------------------------------------------------------#


# get list of column names in Table
cur.execute('''SELECT * FROM {}'''.format(db_table))
db_col_names = [description[0] for description in cur.description]

#-----------------------------------------------------------------------#
#                             Annotation Setup                          #
#-----------------------------------------------------------------------#
annot_file = open(annot_file_name, "r")
annot_dict = {}

# Read header columns into list
header_columns_list = annot_file.readline().split("\t")
header_dict = {}

for i,header in enumerate(header_columns_list):
    header_dict[i] = header

annot_line = annot_file.readline()


#-----------------------------------------------------------------------#
#                         Process Annotations                           #
#-----------------------------------------------------------------------#

while annot_line:
    # Create a dictionary for storing all attributes for this one line
    line_dict = {}
    # Split line since this is a tsv file
    split_line = annot_line.split("\t")
    # Walk through each column value
    for i,element in enumerate(split_line):
        # Save the name of the column/header being processed
        header = header_dict[i].strip()
        # Cleanup extra white space, remove extra quotation marks
        element = element.strip().replace('\"','')

        # If it's the first column (index 0) this is the unique column for matching
        if i == 0:
            unique_header = header
            unique_element = element

        # IF annotation file header is a db column name, retain for annotation
        elif header in db_col_names:
            line_dict[header] = element

    # Check if unique_element is in table
    query = "SELECT * FROM {0} WHERE {1}={2}".format(db_table,
                                                    unique_header,
                                                    "'" + unique_element + "'")

    cur.execute(query)
    fetch_records = cur.fetchall()

    # Check if the record could be found in the database
    if not fetch_records:
        print("Entry not in DB: " + unique_element + ". No annotation is added.", flush = True)
        #raise NCBImetaErrors.ErrorEntryNotInDB(line_strain)
        annot_line = annot_file.readline()
        continue

    # Check if there were multiple hits in the database
    elif len(fetch_records) > 1:
        print("Multiple Matches in DB: " + unique_element + ". No annotation is added.", flush = True)
        #raise NCBImetaErrors.ErrorEntryMultipleMatches(line_strain)
        annot_line = annot_file.readline()
        continue


    # This section allows for dynamic variable creation and column modification
    sql_dynamic_vars = ",".join([header + "=" + "'" + line_dict[header] + "'" for header in line_dict.keys()])
    sql_dynamic_query = "UPDATE {0} SET {1} WHERE {2}={3}".format(db_table,
                                                    sql_dynamic_vars,
                                                    unique_header,
                                                    "'" + unique_element + "'")
    print("Entry " + unique_element + " found in db. " + sql_dynamic_query)
    cur.execute(sql_dynamic_query)

    # Read in the next line
    annot_line = annot_file.readline()


#-----------------------------------------------------------------------#
#                                    Cleanup                            #
#-----------------------------------------------------------------------#
# Commit changes
conn.commit()
print("Closing database: " + db_name, flush = True)
cur.close()
annot_file.close()



"""
NCBImeta Test - Error Classes

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                         Modules and Packages                          #
#-----------------------------------------------------------------------#

import pytest                               # Testing suite
from ncbimeta import NCBImetaErrors         # Utility Functions
import os                                   # Filepath operations
#-----------------------------------------------------------------------#
#                           Test Function                               #
#-----------------------------------------------------------------------#

def test_ErrorOutputDirNotExists(tmpdir):
    '''Test the class ErrorOutputDirNotExists (error when a directory doesn't exist)'''
    tmpdir = tmpdir.strpath
    # Test instantiation
    test_error = NCBImetaErrors.ErrorOutputDirNotExists(tmpdir)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nOutput directory does not exist." + "\n" + "User entered: " + tmpdir)
    assert error_output == error_expect

def test_ErrorAnnotFileNotExists(tmpdir):
    '''Test the class ErrorAnnotFileNotExists (error when an annotation file doesn't exist)'''
    # This file is not created, just a tmp path
    tmpfile = os.path.join(tmpdir.strpath, "tmpfile")
    # Test instantiation
    test_error = NCBImetaErrors.ErrorAnnotFileNotExists(tmpfile)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nFile does not exist." + "\n" + "User entered: --annotfile " + tmpfile)
    assert error_output == error_expect

def test_ErrorTableNotInDB(tmpdir):
    '''Test the class ErrorTableNotInDB (error when a table doesn't exist in a database)'''
    # This file is not created, just a tmp path
    tmpfile = os.path.join(tmpdir.strpath, "tmpfile")
    # Test instantiation
    test_error = NCBImetaErrors.ErrorTableNotInDB(tmpfile)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nThe table does not exist in the database." + "\n" + "Unknown table found: " + tmpfile)
    assert error_output == error_expect

def test_ErrorEntryNotInDB():
    '''Test the class ErrorEntryNotInDB (error when an entry doesn't exist in a database)'''
    # This file is not created, just a tmp path
    test_entry = "TestEntry"
    # Test instantiation
    test_error = NCBImetaErrors.ErrorEntryNotInDB(test_entry)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nThe entry does not exist in the database." + "\n" + "Unknown entry found: " + test_entry)
    assert error_output == error_expect

def test_ErrorEntryMultipleMatches():
    '''Test the class ErrorEntryMultipleMatches (error when their are multiple matching entries in a database)'''
    # This file is not created, just a tmp path
    test_entry = "TestEntry"
    # Test instantiation
    test_error = NCBImetaErrors.ErrorEntryMultipleMatches(test_entry)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nThe entry has multiple matches in the database." + "\n" + "Multiple matches for entry: " + test_entry)
    assert error_output == error_expect

def test_ErrorConfigFileNotExists(tmpdir):
    '''Test the class ErrorAnnotFileNotExists (error when a configuration file doesn't exist)'''
    # This file is not created, just a tmp path
    tmpfile = os.path.join(tmpdir.strpath, "tmpfile")
    # Test instantiation
    test_error = NCBImetaErrors.ErrorConfigFileNotExists(tmpfile)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nConfig file does not exist in the specified location." + "\n" + "Location specified: " + tmpfile)
    assert error_output == error_expect

def test_ErrorColumnsNotUnique():
    '''Test the class ErrorColumnsNotUnique (error when their are non unique columns in a database)'''
    # This file is not created, just a tmp path
    test_column = "TestColumn"
    # Test instantiation
    test_error = NCBImetaErrors.ErrorColumnsNotUnique(test_column)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nThe following columns are not unique in the database:" + "\n" + test_column)
    assert error_output == error_expect

def test_ErrorDBNotExists(tmpdir):
    '''Test the class ErrorDBNotExists (error when a database doesn't exist)'''
    # This file is not created, just a tmp path
    tmpfile = os.path.join(tmpdir.strpath, "tmpfile")
    # Test instantiation
    test_error = NCBImetaErrors.ErrorDBNotExists(tmpfile)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nDatabase does not exist." + "\n" + tmpfile)
    assert error_output == error_expect

def test_ErrorMaxFetchAttemptsExceeded():
    '''Test the class ErrorMaxFetchAttemptsExceeded (error when maximum fetch attempts has been exceeded)'''
    # This file is not created, just a tmp path
    test_ID = '123456789'
    # Test instantiation
    test_error = NCBImetaErrors.ErrorMaxFetchAttemptsExceeded(test_ID)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nThe Maximum number of fetch attempts was exceeded for ID:" + "\n" + test_ID)
    assert error_output == error_expect

def test_ErrorMaxReadAttemptsExceeded():
    '''Test the class ErrorMaxReadAttemptsExceeded (error when maximum read attempts has been exceeded)'''
    # This file is not created, just a tmp path
    test_table = 'TestTable'
    # Test instantiation
    test_error = NCBImetaErrors.ErrorMaxReadAttemptsExceeded(test_table)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nThe Maximum number of read attempts was exceeded for table:" + "\n" + test_table)
    assert error_output == error_expect

def test_ErrorConfigParameter():
    '''Test the class ErrorConfigParameter (error when a configuration file parameter is incorrect)'''
    # This file is not created, just a tmp path
    test_parameter = "TestParameter"
    # Test instantiation
    test_error = NCBImetaErrors.ErrorConfigParameter(test_parameter)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nA parameter name and/or value in the configuration file is set incorrectly:" + "\n" + test_parameter)
    assert error_output == error_expect

def test_ErrorConfigYAMLFormat(tmpdir):
    '''Test the class ErrorConfigYAMLFormat (error when a configuration file is improperly formatted)'''
    # This file is not created, just a tmp path
    tmpfile = os.path.join(tmpdir.strpath, "tmpfile")
    # Test instantiation
    test_error = NCBImetaErrors.ErrorConfigYAMLFormat(tmpfile)
    # Test str representation (error message)
    error_output = str(test_error)
    error_expect = ("\n\nThe configuration file could not be loaded, please confirm that this is a proper YAML file: " + "\n" + tmpfile)
    print("TESTING")
    assert error_output == error_expect



# -*- coding: utf-8 -*-
"""
NCBImeta Utility Functions

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                         Modules and Packages                          #
#-----------------------------------------------------------------------#

import os                               # Filepath operations
from sys import platform as _platform   # Get Platform for OS separator
import sqlite3                          # Database storage and queries
from ncbimeta import NCBImetaErrors     # NCBImeta Error classes
import urllib.error                     # HTTP Error Catching
import time                             # Allow sleeping of processes
from lxml import etree                  # XML Parsing

#-----------------------------------------------------------------------#
#                         Utility Functions                             #
#-----------------------------------------------------------------------#

def check_accessory_dir(output_dir):
    '''
    Check if the accessory directories already exist.

    Parameters:
    output_dir(str): Path to the output directory.
    '''
    if not os.path.exists(os.path.join(output_dir,"log")):
        os.makedirs(os.path.join(output_dir,"log"))
    if not os.path.exists(os.path.join(output_dir,"database")):
        os.makedirs(os.path.join(output_dir,"database"))

def table_exists(db_cur, table_name):
    '''
    Return the result of fetching a table name from a sqlite database to check if it exists.

    Parameters:
    db_cur(Cursor): A sqlite cursor connection.
    table_name(str): Name of a table in the sqite database.
    '''
    query = "SELECT name FROM sqlite_master WHERE type='table' AND name='{}'".format(table_name)
    return db_cur.execute(query).fetchone() is not None

def xml_search(xml_root, search_list, current_tag, column_name, xml_dict):
    '''
    Search xml_root using XPATH for nodes, attributes in search_list and update node_dict.
    In addition to searching, this function also handles escaped XML as text &lt; and &gt;
    characters, as well as CDATA sections.

    Parameters:
    xml_root (ElementTree): xml document as etree object
    search_list (list): list of nodes and attributes in descending hierarchy
    current_tag (str): current tag (or attribute/value) to be searching
    xml_dict (dict): A dictionary to modify and store found values.

    Returns:
    Void. Instead the function mutates the dictionary xml_dict.
    '''
    # Xpath query (could be tag or attribute)
    tag_xpath = ".//"  + current_tag
    #--------------------------------------------------------------------------#
    #                   Recursion Bottom (End of Search List)                  #
    #--------------------------------------------------------------------------#
    if search_list.index(current_tag) == len(search_list) - 1:
        #--- First try the current_tag as attribute ---#
        try:
            fetch_attrib = xml_root.get(current_tag)
            # If successful, add to dictionary
            if fetch_attrib: xml_dict[column_name].append(fetch_attrib)
        except AttributeError: pass
        #--- Next, try the current_tag as a node ---#
        try:
            search_results = xml_root.findall(tag_xpath)
            # Iterate over all possible nodes
            for result in search_results:
                # If the result was a text node
                if result.text:
                    # Remove empty string elements/empty whitespace
                    result_text = result.text.strip()
                    xml_dict[column_name].append(str(result_text))
                # If the result wasn't a text node, but does have child nodes
                elif len(result) > 0:
                    xml_dict[column_name].append(result[0].tag)
        except IndexError: pass
    #--------------------------------------------------------------------------#
    #                                 Ongoing Recursion                        #
    #--------------------------------------------------------------------------#
    else:
        # First try current_tag as node, allowing multiple results
        search_results = xml_root.findall(tag_xpath)
        # We need to know the next tag
        next_tag = search_list[search_list.index(current_tag) + 1]

        #--- No results, first possibility is xml tags have been recoded ---#
        if not search_results:
            # Retry the search with unescaped characters
            xml_root_string_orig = etree.tostring(xml_root)
            open_char = "&lt;"
            close_char = "&gt;"
            xml_root_string = str(xml_root_string_orig).replace(open_char,"<").replace(close_char,">").strip()
            xml_root_string = str(xml_root_string).replace("\\n","").replace("\\t","")
            # Strip off the first 2 char (b') and the final char '
            #print(xml_root_string)
            try:
                xml_root = etree.fromstring(xml_root_string[2:-1])
            # To aggressive, actual greater than/less than signs got replaced in the text
            except etree.XMLSyntaxError:
                xml_root_string = str(xml_root_string_orig).replace("\\n","").replace("\\t","")
                xml_root = etree.fromstring(xml_root_string[2:-1])
            # Now retry the search with the tags fixed up
            search_results = xml_root.findall(tag_xpath)
            # If attributes of interest are present and matching

        #--- See if the info we want exists at this level as attribute values---#
        try:
            if xml_root.get(current_tag) == next_tag:
                xml_dict[column_name].append(xml_root.text)
        except AttributeError: pass

        #--- With multiple results, need to keep searching recursively ---#
        for result in search_results:
            # If there was a result, but not a text node, try attribute
            fetch_attrib = ""
            if not result.text:
                fetch_attrib = result.get(next_tag)
                if fetch_attrib: xml_dict[column_name].append(result.get(next_tag))
            # If there was a text node found
            elif result.text:
                result_text = result.text.strip()
                if result_text:
                    # Check if result contains CDATA
                    first_char = result_text[0]
                    last_char = result_text[-1]
                    if first_char == "<" and last_char == ">":
                        # Reformat CDATA as complete XML
                        result_text = ("<" + current_tag + ">" + result_text + "</" + current_tag + ">")
                        result = etree.fromstring(result_text)

            if not fetch_attrib:
                # If an attribute wasn't succesfully fetched, need recursion
                xml_search(result, search_list, next_tag, column_name, xml_dict)

def HTTPErrorCatch(http_method, max_fetch_attempts, sleep_time, **kwargs):
    '''
    Return result of http_method and check if an HTTP Error is generated

    Parameters:
    http_method (function): An http record-fetching or searching method.
    max_fetch_attempts (int): Maximum number of tries for fetching a record_dict.
    sleep_time (float): Number of seconds to wait in between fetch read_attempts.
    kwargs(dict): keyword arguments for the http_method function.
    '''
    # Attemp the http_method function, wrapped in HTTP error checking
    ID_handle_retrieved = False
    fetch_attempts = 0
    while not ID_handle_retrieved and fetch_attempts < max_fetch_attempts:
        try:
            ID_handle = http_method(**kwargs)
            ID_handle_retrieved = True
        # HTTP Errors
        except urllib.error.HTTPError as error:
            # Error code 429: Too Many Requests
            if error.code == 429:
                fetch_attempts += 1
                print("HTTP Error " + str(error.code) + ": " + str(error.reason))
                print("Fetch Attempt: " + str(fetch_attempts) + "/" + str(max_fetch_attempts))
                print("Sleeping for " + str(sleep_time) + " seconds before retrying.")
                time.sleep(sleep_time)
            # General HTTP Error Code, non specific
            else:
                fetch_attempts += 1
                print("HTTP Error " + str(error.code) + ": " + str(error.reason))
                print("Fetch Attempt: " + str(fetch_attempts) + "/" + str(max_fetch_attempts))
                print("Retrying record fetching.")
        # URL Errors
        except urllib.error.URLError as error:
            fetch_attempts += 1
            print("URL Error: " + str(error.reason))
            print("Fetch Attempt: " + str(fetch_attempts) + "/" + str(max_fetch_attempts))
            print("Retrying record fetching.")


        # If the maximum number of fetch attempts has been exceeded
        if fetch_attempts == max_fetch_attempts and not ID_handle_retrieved:
            raise NCBImetaErrors.ErrorMaxFetchAttemptsExceeded(ID)

    return ID_handle



"""
NCBImeta Test - Export

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                         Modules and Packages                          #
#-----------------------------------------------------------------------#

import pytest                             # Testing suite
from ncbimeta import *                    # Main Program
import os                                 # Filepath operations
import test.test_ncbimeta                 # Run the program to create test db
import subprocess                         # Execute CLI/Shell

#-----------------------------------------------------------------------#
#                           Test Function                               #
#-----------------------------------------------------------------------#

def test_export_run():
    '''Test the NCBImetaExport application for run completion'''
    # Use the test database
    test_db = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test.sqlite")
    test_output_dir = os.path.dirname(os.path.abspath(__file__))
    # If the test_db doesn't alread exist, run the test cmd from test_ncbimeta
    if not os.path.exists(test_db): test_ncbimeta.test_ncbimeta_run()
    test_cmd = ("ncbimeta/NCBImetaExport.py --database " + test_db +
                " --outputdir " + test_output_dir)

    # test NCBImetaExport through a subprocess
    returned_value = subprocess.call(test_cmd, shell=True)
    # If it returns a non-zero value, it failed
    assert returned_value == 0

def test_export_assemblyvalues(assembly_table_data):
    '''
    Test the integrity of the Assembly table values based on expected values.

    Parameters:
    assembly_table_data (fixture): Dict fixture of Assembly table data from conftest.py
    '''
    # Setup the assembly table file
    test_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test_Assembly.txt")
    test_file = open(test_filename,'r')
    # Retrieve the headers and fields
    test_column_list = test_file.readline().strip('\n').split("\t")
    # Add empty tab on the end for empty comment field
    test_metadata_list = (test_file.readline().strip('\n') + "\t" + "").split("\t")
    # Populate the dict with data
    test_dict = {}

    for i in range(0,len(test_column_list)):
        key = test_column_list[i]
        value = test_metadata_list[i]
        test_dict[key] = value
    # Test whether the values are as expected
    assert test_dict == assembly_table_data
    #Cleanup
    test_file.close()

def test_export_bioprojectvalues(bioproject_table_data):
    '''
    Test the integrity of the BioProject table values based on expected values.

    Parameters:
    bioproject_table_data (fixture): Dict fixture of BioProject table data from conftest.py
    '''
    # Setup the assembly table file
    test_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test_BioProject.txt")
    test_file = open(test_filename,'r')
    # Retrieve the headers and fields
    test_column_list = test_file.readline().strip('\n').split("\t")
    test_metadata_list = (test_file.readline().strip('\n') + "\t" + "").split("\t")
    # Populate the dict with data
    test_dict = {}
    for i in range(0,len(test_column_list)):
        key = test_column_list[i]
        value = test_metadata_list[i]
        test_dict[key] = value
    # Test whether the values are as expected
    assert test_dict == bioproject_table_data
    #Cleanup
    test_file.close()

def test_export_biosamplevalues(biosample_table_data):
    '''
    Test the integrity of the BioSample table values based on expected values.

    Parameters:
    biosample_table_data (fixture): Dict fixture of BioSample table data from conftest.py
    '''
    # Setup the assembly table file
    test_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test_BioSample.txt")
    test_file = open(test_filename,'r')
    # Retrieve the headers and fields
    test_column_list = test_file.readline().strip('\n').split("\t")
    test_metadata_list = (test_file.readline().strip('\n') + "\t" + "").split("\t")
    # Populate the dict with data
    test_dict = {}
    for i in range(0,len(test_column_list)):
        key = test_column_list[i]
        value = test_metadata_list[i]
        test_dict[key] = value
    # Test whether the values are as expected
    assert test_dict == biosample_table_data
    #Cleanup
    test_file.close()

def test_export_nucleotidevalues(nucleotide_table_data):
    '''
    Test the integrity of the Nucleotide table values based on expected values.

    Parameters:
    nucleotide_table_data (fixture): Dict fixture of Nucleotide table data from conftest.py
    '''
    # Setup the assembly table file
    test_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test_Nucleotide.txt")
    test_file = open(test_filename,'r')
    # Retrieve the headers and fields
    test_column_list = test_file.readline().strip('\n').split("\t")

    # Populate the dict with data
    test_dict = {}
    read_line = test_file.readline().strip('\n')
    while read_line:
        test_metadata_list = (read_line + "\t" + "").split("\t")
        for i in range(0,len(test_column_list)):
            key = test_column_list[i]
            value = test_metadata_list[i]
            # Check if key is in dict, if not create as list
            if key not in test_dict:
                test_dict[key] = [value]
            # Otherwise, append to it
            else:
                test_dict[key].append(value)

        read_line = test_file.readline().strip('\n')

    # Test whether the values are as expected
    assert test_dict == nucleotide_table_data
    #Cleanup
    test_file.close()

def test_export_pubmedvalues(pubmed_table_data):
    '''
    Test the integrity of the Pubmed table values based on expected values.

    Parameters:
    pubmed_table_data (fixture): Dict fixture of Pubmed table data from conftest.py
    '''
    # Setup the assembly table file
    test_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test_Pubmed.txt")
    test_file = open(test_filename,'r')
    # Retrieve the headers and fields
    test_column_list = test_file.readline().strip('\n').split("\t")
    # Add an extra empty string for the empty comment
    test_metadata_list = (test_file.readline().strip('\n')  + "\t" + "").split("\t")
    # Populate the dict with data
    test_dict = {}
    for i in range(0,len(test_column_list)):
        key = test_column_list[i]
        value = test_metadata_list[i]
        test_dict[key] = value

    # Test whether the values are as expected
    assert test_dict == pubmed_table_data
    #Cleanup
    test_file.close()

def test_export_sravalues(sra_table_data):
    '''
    Test the integrity of the SRA table values based on expected values.

    Parameters:
    sra_table_data (fixture): Dict fixture of SRA table data from conftest.py
    '''
    # Setup the assembly table file
    test_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test_SRA.txt")
    test_file = open(test_filename,'r')
    # Retrieve the headers and fields
    test_column_list = test_file.readline().strip('\n').split("\t")

    # Populate the dict with data
    test_dict = {}
    read_line = test_file.readline().strip('\n')
    while read_line:
        test_metadata_list = (read_line + "\t" + "").split("\t")
        for i in range(0,len(test_column_list)):
            key = test_column_list[i]
            value = test_metadata_list[i]
            # Check if key is in dict, if not create as list
            if key not in test_dict:
                test_dict[key] = [value]
            # Otherwise, append to it
            else:
                test_dict[key].append(value)

        read_line = test_file.readline().strip('\n')

    # Test whether the values are as expected
    assert test_dict == sra_table_data
    #Cleanup
    test_file.close()

# def test_export_mastervalues(assembly_table_data,
#                                 bioproject_table_data,
#                                 biosample_table_data,
#                                 nucleotide_table_data,
#                                 pubmed_table_data,
#                                 sra_table_data):
#     '''
#     Test the integrity of the Master table values based on expected values.
#
#     Parameters:
#     biosample_table_data (fixture): Dict fixture of BioSample table data from conftest.py
#     '''
#     # Setup the assembly table file
#     test_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test_Master.txt")
#     test_file = open(test_filename,'r')
#     # Retrieve the headers and fields
#     test_column_list = test_file.readline().strip('\n').split("\t")
#     test_metadata_list = test_file.readline().strip('\n').split("\t")
#     # Populate the dict with data
#     test_dict = {}
#     for i in range(0,len(test_column_list)):
#         key = test_column_list[i]
#         value = test_metadata_list[i]
#         test_dict[key] = value
#
#
#     # Construct the master_table_data
#     master_table_data = {'id':'1'}
#     # Nucleotide and sra are multi-row tables
#     for record in nucleotide_table_data:
#         # Concatenate the values
#         unique_value_list = list(set(nucleotide_table_data[record]))
#         nucleotide_table_data[record] = ";".join(unique_value_list)
#
#     for record in sra_table_data:
#         # Concatenate the values
#         unique_value_list = list(set(sra_table_data[record]))
#         sra_table_data[record] = ";".join(unique_value_list)
#
#     for table in assembly_table_data, bioproject_table_data, biosample_table_data, nucleotide_table_data, pubmed_table_data, sra_table_data:
#         table.pop('id', None)
#         master_table_data.update(table)
#
#     # Test whether the values are as expected
#     assert test_dict == master_table_data
#     #Cleanup
#     test_file.close()



import setuptools

with open("README.md", 'r') as f:
    long_description = f.read()

with open("requirements.txt", 'r') as r:
    require_list = r.read().strip().split("\n")

setuptools.setup(
   name='NCBImeta',
   version='0.6.0',
   description='Efficient and comprehensive metadata acquisition from the NCBI databases (includes SRA).',
   python_requires='>=3',
   license="MIT",
   long_description=long_description,
   long_description_content_type='text/markdown',
   author='Katherine Eaton',
   author_email='ktmeaton@gmail.com',
   url='https://ktmeaton.github.io/NCBImeta/',
   #packages=setuptools.find_packages(),
   packages=['ncbimeta'],
   classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
   ],
   install_requires=require_list, #external packages as dependencies
   scripts=[
      'ncbimeta/NCBImeta.py',
      'ncbimeta/NCBImetaExport.py',
      'ncbimeta/NCBImetaJoin.py',
      'ncbimeta/NCBImetaAnnotateReplace.py',
      'ncbimeta/NCBImetaAnnotateConcatenate.py']
)



"""
NCBImeta Test - Main CLI (NCBImeta)

@author: Katherine Eaton
"""

#-----------------------------------------------------------------------#
#                         Modules and Packages                          #
#-----------------------------------------------------------------------#

import pytest                           # Testing suite
from ncbimeta import *                  # Main Program Import
from ncbimeta import NCBImetaErrors     # Error classes
import yaml                             # YAML config file parsing
import os                               # Filepath operations
import subprocess                       # Execute CLI/Shell

#-----------------------------------------------------------------------#
#                           Test Function                               #
#-----------------------------------------------------------------------#

def test_ncbimeta_run():
    '''Test the NCBImeta application for run completion'''
    # User the stripped down testing config file
    config_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test.yaml")
    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file
    # test the main NCBImeta.py through a subprocess
    returned_value = subprocess.call(test_cmd, shell=True)
    # If it returns a non-zero value, it failed
    assert returned_value == 0

def test_ncbimeta_noflatmode():
    '''Test the NCBImeta application for run completion'''
    # User the stripped down testing config file
    config_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test.yaml")
    test_cmd = "ncbimeta/NCBImeta.py --config " + config_file
    # test the main NCBImeta.py through a subprocess
    returned_value = subprocess.call(test_cmd, shell=True)
    # If it returns a non-zero value, it failed
    assert returned_value == 0

def test_ncbimeta_ConfigFileNotExists(tmpdir):
    '''Test for error catching when a config file does not exist'''
    # This file is not created, just a tmp path
    config_file = os.path.join(tmpdir.strpath, "tmpfile")
    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigFileNotExists" in str(e.output)

def test_ncbimeta_ConfigFileNotYAML(tmpdir):
    '''Test for error catching when a config file has improper format'''
    # User the empty config file
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'a').close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigYAMLFormat" in str(e.output)
    os.remove(config_file_name)

def test_ncbimeta_ErrorConfigParameter_OUTPUT_DIR(tmpdir):
    '''Test for error catching when a config file has improper parameters'''
    # Prep the config file contents
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'w')
    config_file.write('OUTPUT_DIRR: output')
    config_file.close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigParameter" in str(e.output)
    # Cleanup
    os.remove(config_file_name)

def test_ncbimeta_ErrorConfigParameter_EMAIL(tmpdir):
    '''Test for error catching when a config file has improper parameters'''
    # Prep the config file contents
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'w')
    config_file.write('OUTPUT_DIR: output' + '\n' +
                      'EMAILL: email')
    config_file.close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigParameter" in str(e.output)
    # Cleanup
    os.remove(config_file_name)

def test_ncbimeta_ErrorConfigParameter_API_KEY(tmpdir):
    '''Test for error catching when a config file has improper parameters'''
    # Prep the config file contents
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'w')
    config_file.write('OUTPUT_DIR: output' + '\n' +
                      'EMAIL: email' + '\n' +
                      'API_KEYY: api_key')
    config_file.close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigParameter" in str(e.output)
    # Cleanup
    os.remove(config_file_name)

def test_ncbimeta_ErrorConfigParameter_FORCE_PAUSE_SECONDS(tmpdir):
    '''Test for error catching when a config file has improper parameters'''
    # Prep the config file contents
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'w')
    config_file.write('OUTPUT_DIR: output' + '\n' +
                      'EMAIL: email' + '\n' +
                      'API_KEY: api_key' + '\n' +
                      'FORCE_PAUSE_SECONDSS: force_pause_seconds')
    config_file.close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigParameter" in str(e.output)
    # Cleanup
    os.remove(config_file_name)

def test_ncbimeta_ErrorConfigParameter_DATABASE(tmpdir):
    '''Test for error catching when a config file has improper parameters'''
    # Prep the config file contents
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'w')
    config_file.write('OUTPUT_DIR: output' + '\n' +
                      'EMAIL: email' + '\n' +
                      'API_KEY: api_key' + '\n' +
                      'FORCE_PAUSE_SECONDS: force_pause_seconds' + '\n' +
                      'DATABASEE: database')
    config_file.close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigParameter" in str(e.output)
    # Cleanup
    os.remove(config_file_name)

def test_ncbimeta_ErrorConfigParameter_TABLES(tmpdir):
    '''Test for error catching when a config file has improper parameters'''
    # Prep the config file contents
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'w')
    config_file.write('OUTPUT_DIR: output' + '\n' +
                      'EMAIL: email' + '\n' +
                      'API_KEY: api_key' + '\n' +
                      'FORCE_PAUSE_SECONDS: force_pause_seconds' + '\n' +
                      'DATABASE: database' + '\n' +
                      'TABLESS: tables')
    config_file.close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigParameter" in str(e.output)
    # Cleanup
    os.remove(config_file_name)

def test_ncbimeta_ErrorConfigParameter_SEARCH_TERMS(tmpdir):
    '''Test for error catching when a config file has improper parameters'''
    # Prep the config file contents
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'w')
    config_file.write('OUTPUT_DIR: output' + '\n' +
                      'EMAIL: email' + '\n' +
                      'API_KEY: api_key' + '\n' +
                      'FORCE_PAUSE_SECONDS: force_pause_seconds' + '\n' +
                      'DATABASE: database' + '\n' +
                      'TABLES: tables' + '\n' +
                      'SEARCH_TERMSS: search_terms')
    config_file.close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigParameter" in str(e.output)
    # Cleanup
    os.remove(config_file_name)

def test_ncbimeta_ErrorConfigParameter_TABLE_COLUMNS(tmpdir):
    '''Test for error catching when a config file has improper parameters'''
    # Prep the config file contents
    tmpdir = tmpdir.strpath
    config_file_name = os.path.join(tmpdir,'tmp_config.yaml')
    config_file = open(config_file_name, 'w')
    config_file.write('OUTPUT_DIR: output' + '\n' +
                      'EMAIL: email' + '\n' +
                      'API_KEY: api_key' + '\n' +
                      'FORCE_PAUSE_SECONDS: force_pause_seconds' + '\n' +
                      'DATABASE: database' + '\n' +
                      'TABLES: tables' + '\n' +
                      'SEARCH_TERMS: search_terms' + '\n' +
                      'TABLE_COLUMNSS: table_columns')
    config_file.close()

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file_name
    # Use a try-catch with the sub-process of module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorConfigParameter" in str(e.output)
    # Cleanup
    os.remove(config_file_name)

def test_ncbimeta_OutputDirNotExists(tmpdir):
    '''Test for error catching when the output dir does not exist'''
    tmpdir = tmpdir.strpath
    # Prep files and directories
    src_config = os.path.join(os.path.dirname(os.path.abspath(__file__)),"test.yaml")
    dst_dir = os.path.join(tmpdir, "tmp-output-test")
    config_file = os.path.join(tmpdir,'tmp_config.yaml')

    # Read in the file
    with open(src_config, 'r') as file: filedata = file.read()
    # Replace the target data
    filedata = filedata.replace('OUTPUT_DIR : test', 'OUTPUT_DIR : ' + dst_dir )
    # Write to the destination file
    with open(config_file, 'a') as file: file.write(filedata)

    test_cmd = "ncbimeta/NCBImeta.py --flat --config " + config_file
    # Use a try-catch with the sub-process module
    try:
        subprocess.check_output(test_cmd,
                                shell=True,
                                stderr=subprocess.STDOUT)
        # If an exception isn't raise then the assertion fails (0)
        assert 0
    except subprocess.CalledProcessError as e:
        # If the right error is raise, it's Error Class is in the output
        assert "ErrorOutputDirNotExists" in str(e.output)
    # Cleanup
    os.remove(config_file)
